part of fisher;

abstract class PointType {
  static const String 
    CIRCLE = "_o",
    SQUARE = "_0",
    DIAMOND = "_D",
    TRIANGLE_UP = "_^",
    TRIANGLE_LEFT = "_<",
    TRIANGLE_RIGHT = "_>",
    TRIANGLE_DOWN = "_v",
    PLUS = "_+",
    TIMES = "_x"
  ; 
}

/// A representation of a plot.
class Plot {
  CanvasElement canvas, buffer, layer;
  CanvasRenderingContext2D context, bufferContext, layerContext;
  Map properties;
  List<num> xRange, yRange, drawXRange, drawYRange;
  
/// Constructs an instance of [Plot]. 
/// 
/// Takes in the named arguments:
///  * [width] _image width in pixels_
///  * [height] _image height in pixels_
///  * [xRange] _sub-domain the image covers_
///  * [yRange] _sub-codomain the image covers_
///  * [drawXRange] _sub-domain visible in the image_
///  * [drawYRange] _sub-codomain visible in the image_
/// 
/// The constructed object will have a [CanvasElement] property
/// called [canvas] that can be added to an HTML element.
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot(
///       width: 600,
///       height: 600,
///       xRange: [-10, 10],
///       yRange: [-10, 10],
///       drawXRange: [-8, 9],
///       drawYRange: [-8, 8]
///     );
/// 
/// ...
  Plot({
    int width: 600, 
    int height,
    List<num> xRange: const [-10.0, 10.0],
    List<num> yRange: const [-10.0, 10.0],
    List<num> drawXRange,
    List<num> drawYRange
  }) {
    if (height == null) height = (width * 1).round();
    
    canvas = new CanvasElement(width: width, height: height);
    context = canvas.getContext("2d");
    
    buffer = new CanvasElement(width: width, height: height);
    bufferContext = buffer.getContext("2d");
        
    layer = new CanvasElement(width: width, height: height);
    layerContext = layer.getContext("2d");
    layerContext..lineCap = "round"..lineJoin = "round";
    
    if (drawXRange == null) drawXRange = new List<num>.from(xRange); 
    if (drawYRange == null) drawYRange = new List<num>.from(yRange); 
    this
      ..xRange = xRange
      ..yRange = yRange
      ..drawXRange = drawXRange
      ..drawYRange = drawYRange
    ;
  }
  
  _transformX(num x) => 
    ((x - xRange[0]) / (xRange[1] - xRange[0])
      * canvas.width);
  
  _transformY(num y) =>
    (canvas.height - (y - yRange[0]) / (yRange[1] - yRange[0])
      * canvas.height);
  
  List<num> _autoscale(
    num max, 
    num min, 
    List<num> bounds, 
    [int pow10]
  ) {
    const idealMin = 2, idealMax = 12;
    if (pow10 == null) 
      pow10 = (Math.log(max - min) / Math.LN10).floor();
    num scale = Math.pow(10, pow10);
    List<num> axis = sequence(
      ((min - (max - min)) / scale).floor() * scale,
      ((max + (max - min)) / scale).floor() * scale,
      by: scale
    ).where((num x) => x > bounds[0] && x < bounds[1])
     .toList();
    if (axis.length < idealMin) return _autoscale(max, min, bounds, pow10 - 1);
    else if (axis.length > idealMax) return _autoscale(max, min, bounds, pow10 + 1);
    else return axis;
  }
  
  _setUpPlotRegion({
    num minX: -10.0,
    num maxX: 10.0,
    num minY: -10.0,
    num maxY: 10.0,
    num top,
    num topBuffer: 0.1,
    num bottom, 
    num bottomBuffer: 0.1,
    num left,
    num leftBuffer: 0.1,
    num right,
    num rightBuffer: 0.1,
    String title,
    String xLabel,
    String yLabel,
    bool markX: false,
    bool markY: false,
    bool border: true,
    bool summary: false
  }) {
    if (top == null) top = summary ? 0.1 : 0.3;
    if (bottom == null) bottom = summary ? 0.1 : 0.2;
    if (left == null) left = summary ? 0.1 : 0.2;
    if (right == null) right = summary ? 0.1 : 0.1;
    
    num
      width = maxX - minX,     // width of visible rectangle
      height = maxY - minY     // height of visible rectangle
    ;
    
    xRange = [
      minX - (left + leftBuffer) * width,
      maxX + (right + rightBuffer) * width
    ];
    
    drawXRange = [
      minX - leftBuffer * width,
      maxX + rightBuffer * width
    ];
    
    yRange = [
      minY - (bottom + bottomBuffer) * height,
      maxY + (top + topBuffer) * height
    ];
    
    drawYRange = [
      minY - bottomBuffer * height,
      maxY + topBuffer * height
    ];
      
    // Add title
    if (title != null) labels(
      x: (minX + maxX) / 2,
      y: maxY + (topBuffer + 0.5 * top) * height,
      labels: title,
      fontSize: canvas.height / 25
    );
      
    // Add x-axis label
    if (xLabel != null) labels(
      x: (minX + maxX) / 2,
      y: minY - (bottomBuffer + 0.75 * bottom) * height,
      labels: xLabel,
      fontSize: canvas.height / 35
    );
      
    // Add y-axis label
    if (yLabel != null) labels(
      x: minX - (leftBuffer + 0.75 * left) * width,
      y: (minY + maxY) / 2,
      labels: yLabel,
      fontSize: canvas.height / 35,
      theta: Math.PI / 2
    );
     
    List<num> 
      xAxis = _autoscale(maxX, minX, drawXRange),
      yAxis = _autoscale(maxY, minY, drawYRange)
    ;
    
    var cleanLabels = (List<num> numList) => 
      numList.map(
        (num a) {
          String label = (double.parse(a.toStringAsPrecision(5))).toString();
          
          if (label.substring(label.length - 2) == ".0")
            label = label.substring(0, label.length - 2);//label.replaceFirst(new RegExp(r'\.0*'), "");
          
          return label;
        }
      )
    .toList();
      
    if (markX) {
      marks(x: xAxis, y: drawYRange[0], up: (maxY - minY) / 50);
      // x-axis number line
      labels(
        x: xAxis.map((x) => _transformX(x)).toList(), 
        y: _transformY(drawYRange[0]) + canvas.height / 50, 
        labels: cleanLabels(xAxis),
        color: "gray",
        inDrawRegion: false,
        scaled: false
      );
    }
    
    if (markY) {
      marks(y: yAxis, x: drawXRange[0], right: (maxX - minX) / 50);
      // y-axis number line
      labels(
        x: _transformX(drawXRange[0]) - canvas.width / 100,
        y: yAxis.map((y) => _transformY(y)).toList(),  
        labels: cleanLabels(yAxis),
        color: "gray",
        align: "right",
        inDrawRegion: false,
        scaled: false
      );
    }
    
    if (border) {
      linesVertical(
        x: drawXRange,
        color: "gray"
      );
      linesHorizontal(
        y: drawYRange,
        color: "black"
      );
    }
  }
  
  _drawRegion([
    bool inDrawRegion = true
  ]) {
    
    bufferContext.clearRect(0, 0, buffer.width, buffer.height);
    layerContext.clearRect(0, 0, buffer.width, buffer.height);
    
    bufferContext
      ..globalCompositeOperation = "source-over"
      ..fillStyle = "lightgray"
      ..fillRect(
          inDrawRegion ? _transformX(drawXRange[0]) : 0, 
          inDrawRegion ? _transformY(drawYRange[0]) : 0,
          inDrawRegion ? 
            _transformX(drawXRange[1]) - _transformX(drawXRange[0]) : 
              buffer.width,
          inDrawRegion ? 
            _transformY(drawYRange[1]) - _transformY(drawYRange[0]) :
              buffer.height
        )
      ..globalCompositeOperation = "source-in"
      
    ;

  }
  
/// Adds labels to the plot. 
/// 
/// The argument [labels] may be a [String] or a [List]; 
/// [font], [fontSize] and [color] may be legal CSS expressions or [List]s
/// of legal CSS expressions, in which case they are iterated along with 
/// [labels] to generate labels with different appearances. Similarly, [x] 
/// and [y], which determine label positions, and [theta], which sets 
/// counter-clockwise rotation in radian, may be a [num] or a [List]. The 
/// argument [inDrawRegion] determines whether the labels will only be 
/// visible in the draw region determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.labels(
///       x: [-1, 3, 2],
///       y: [0, -2, 5],
///       labels: ["A", "B", "C"],
///       font: "Courier New",
///       fontSize: [12, 20, 28],
///       color: ["orange", "purple", "blue"]
///     );
///     
/// ...
  void labels({
    var x,
    var y,
    var labels,
    var font: "Arial Rounded MT Bold",
    var fontSize,
    var color: "black",
    var theta: 0,
    var align: "centre",
    bool inDrawRegion: false,
    bool scaled: true
  }) {
    if (fontSize == null) fontSize = (canvas.height / 35).round();
    List<num>
      xData = _extractNumList(x),
      yData = _extractNumList(y),
      fontSizeData = _extractNumList(fontSize),
      thetaData =_extractNumList(theta)
    ;
    
    List 
      labelsData = _extractStrList(labels),
      fontData = _extractStrList(font),
      colorData = _extractStrList(color)
    ;
      
    _drawRegion(inDrawRegion);
    
    int length = Math.max(xData.length, yData.length);
    for (int i = 0; i < length; i++) {
      
      num thetaTransformed = thetaData[i % thetaData.length];
      /*{
        num 
          spanX = Math.cos(thetaTransformed) * _transformX(1),
          spanY = Math.sin(thetaTransformed) * _transformY(1)
        ;
        thetaTransformed = Math.atan2(spanX, spanY);
      }*/
      
      layerContext             
        ..font = "${
          fontSizeData[i % fontSizeData.length]}px ${
            fontData[i % fontData.length]}"
      ;
      
      num 
        textWidth = layerContext
          .measureText(labelsData[i % labelsData.length].toString())
          .width,
        tx = scaled ? _transformX(xData[i % xData.length]) : 
          xData[i % xData.length],
        ty = scaled ? _transformY(yData[i % yData.length]) :
          yData[i % yData.length]
      ;
      
      layerContext
        ..save()
        
        ..translate(
          tx, 
          ty
        )
        ..rotate(-thetaTransformed)
        
        ..fillStyle = colorData[i % colorData.length]
        ..fillText(
          labelsData[i % labelsData.length].toString(), 
          align == "centre" ? -textWidth / 2 :
            align == "left" ? 0 :
              -textWidth, 
          fontSizeData[i % fontSizeData.length] * 0.25
        )
        ..restore()
      ;
    }
      
    bufferContext
      ..drawImage(layer, 0, 0)
    ;
      
    context.drawImage(buffer, 0, 0);
    
  }
  
/// Adds a set of lines to the plot. 
/// 
/// The arguments [width] and [color] may be legal CSS expressions or 
/// [List]s of legal CSS expressions, in which case they are iterated 
/// along with [x] and [y], which determine line end-points. The argument 
/// [inDrawRegion] determines whether the lines will only be visible in the
/// draw region determined by [drawXRange] and [drawYRange].
///  
/// ...
/// Example:
///  
///     var plot = new Plot();
///     plot.lines(
///       x: [0, 1],
///       y: [0, 8],
///       color: "red"
///     );
///      
///  ...
  lines({
    var x,
    var y,
    var width: 1,
    String color: "gray",
    bool inDrawRegion: true
  }) {
    List<num> 
      xData = _extractNumList(x), 
      yData = _extractNumList(y),
      lineWidthData = _extractNumList(width)
    ;
    
    _drawRegion(inDrawRegion);
    
    int length = Math.max(xData.length, yData.length);
    for (int i = 0; i < length - 1; i++) {
      
      layerContext
        ..lineWidth = lineWidthData[i % lineWidthData.length]
        ..strokeStyle = "black"
        ..beginPath()
        ..moveTo(
            _transformX(xData[i % xData.length]), 
            _transformY(yData[i % yData.length])
          )
        ..lineTo(
            _transformX(xData[(i + 1) % xData.length]), 
            _transformY(yData[(i + 1) % yData.length])
          )
        ..stroke()
        ..closePath()
      ;
    }
    bufferContext
        ..fillStyle = color
        ..fillRect(0, 0, canvas.width, canvas.height)
        ..globalCompositeOperation = "destination-in"
        ..drawImage(layer, 0, 0)
    ;
    
    context.drawImage(buffer, 0, 0);
    
  }
  
/// Adds a set of marks to the plot. 
/// 
/// The parameter [width] and [color] may be legal CSS expressions or [List]s 
/// of legal CSS expressions, in which case they are iterated along with [x] 
/// and [y], which determine line end-points. Similarly, [up] and [right] may 
/// be [num]s or [List]s of and determine the length and direction of the 
/// marks (use negative values for leftward and downward marks). The argument
/// [inDrawRegion] determines whether the marks will only be visible in the 
/// draw region determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.marks(
///       x: [0, 1, 2, 3, 4],
///       y: 0,
///       up: 0.1
///     );
///     
/// ...
  marks({
    var x, 
    var y,
    var up: 0,
    var right: 0,
    var color: "gray",
    var width: 1,
    bool inDrawRegion: true
  }) {
    List<num>
      xData = _extractNumList(x),
      yData = _extractNumList(y),
      upData = _extractNumList(up),
      rightData = _extractNumList(right),
      lineWidthData = _extractNumList(width)
    ;
    List<String>
      linecolorData = _extractStrList(color)
    ;
    
    int length = Math.max(xData.length, yData.length);
    for (int i = 0; i < length; i++) {
      num 
        x0 = xData[i % xData.length],
        x1 = xData[i % xData.length] 
          + rightData[i % rightData.length],
        y0 = yData[i % yData.length],
        y1 = yData[i % yData.length] 
          + upData[i % upData.length]
      ;
      lines(
        x: [x0, x1],
        y: [y0, y1],
        color: color,
        width: width,
        inDrawRegion: inDrawRegion
      );
    }
    
  }
  
/// Adds a set of vertical lines to the plot. 
/// 
/// Used similarly to [lines].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.linesVertical(x: [0, 1]);
///     
/// ...
  linesVertical({
    var x,
    var color: "rgba(0,0,0,0.3)",
    var width: 1,
    bool inDrawRegion: true
  }) {
    List<num> 
      xData = _extractNumList(x),
      lineWidthData = _extractNumList(width)
    ;
    List<String>
      linecolorData = _extractStrList(color)
    ;
    for (int i = 0; i < xData.length; i++) {
      lines(
        x: [xData[i], xData[i]],
        y: [yRange[0] - 1, yRange[1] + 1],
        color: linecolorData[i % linecolorData.length],
        width: lineWidthData[i % lineWidthData.length],
        inDrawRegion: inDrawRegion
      );
    }
  }
  
/// Adds a set of horizontal lines to the plot. 
/// 
/// Used similarly to [lines].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.linesHorizontal(x: [0, 1]);
///     
/// ...
  linesHorizontal({
    var y,
    var color: "rgba(0,0,0,0.3)",
    var width: 1,
    bool inDrawRegion: true
  }) {
    List<num> 
      yData = _extractNumList(y),
      lineWidthData = _extractNumList(width)
    ;
    List<String>
      linecolorData = _extractStrList(color)
    ;
    for (int i = 0; i < yData.length; i++) {
      lines(
        x: [xRange[0] - 1, xRange[1] + 1],
        y: [yData[i], yData[i]],
        color: linecolorData[i % linecolorData.length],
        width: lineWidthData[i % lineWidthData.length],
        inDrawRegion: inDrawRegion
      );
    } 
  }
  
/// Adds a grid to the plot.
/// 
/// The arguments [xMajor], [xMinor], [yMajor] and [yMinor]
/// specify the positions of vertical and horizontal lines;
/// the color and width settings can be legal CSS expressions
/// or [List]s of legal CSS expressions.
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.grid(
///       xMajor: sequence(-10, 11, by: 1),
///       xMinor: sequence(-10, 10.5, by: 0.5),
///       yMajor: sequence(-10, 11, by: 1),
///       yMinor: sequence(-10, 10.5, by: 0.5)
///     );
///     
/// ...
  grid({
    var xMajor,
    var xMajorcolor: "rgba(50, 100, 255, 0.1)",
    var xMajorWidth: 1,
    var xMinor,
    var xMinorcolor: "rgba(50, 100, 255, 0.05)",
    var xMinorWidth: 1,
    var yMajor,
    var yMajorcolor: "rgba(50, 100, 255, 0.1)",
    var yMajorWidth: 1,
    var yMinor,
    var yMinorcolor: "rgba(50, 100, 255, 0.05)",
    var yMinorWidth: 1,
    bool inDrawRegion: true
  }){
    
    if (xMajor == null) 
      xMajor = _autoscale(xRange[1], xRange[0], xRange);
    if (xMinor == null) {
      num d = (xMajor[1] - xMajor[0]) / 2;
      xMinor = xMajor.map((x) => x - d)
        .toList()..add(xMajor.last + d);
    }

    if (yMajor == null) 
      yMajor = _autoscale(yRange[1], yRange[0], yRange);
    if (yMinor == null) {
      num d = (yMajor[1] - yMajor[0]) / 2;
      yMinor = yMajor.map((x) => x - d)
        .toList()..add(yMajor.last + d);
    }
    
    linesVertical(
     x: xMinor,
     color: xMinorcolor,
     inDrawRegion: inDrawRegion
    );
    
    linesHorizontal(
      y: yMinor,
      color: yMinorcolor,
     inDrawRegion: inDrawRegion
    );
    
    linesVertical(
      x: xMajor,
      color: xMajorcolor,
      inDrawRegion: inDrawRegion
    );
    
    linesHorizontal(
      y: yMajor,
      color: yMajorcolor,
      inDrawRegion: inDrawRegion
    ); 
    
  }
  
/// Adds the axes to the plot.  
  axes({
    color: "rgba(0,0,0,0.4)",
    width: 1
  }) {
    linesVertical(x: 0, color: color, width: width);
    linesHorizontal(y: 0, color: color, width: width);
  }
  
/// Adds a radial pattern to the plot that can be useful for
/// illustrating polar or parametric functions.
/// 
/// The parameters [rMajor] and [rMinor] specify the radii of the circles;
/// the color and width settings can be legal CSS expressions or [List]s of
/// legal CSS expressions.
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.radial(); 
///     
/// ...
  radial({
    var rMajor,
    var rMajorcolor: "rgba(50, 100, 255, 0.1)",
    var rMajorWidth: 1,
    var rMinor,
    var rMinorcolor: "rgba(50, 100, 255, 0.05)",
    var rMinorWidth: 1
  }) {
    if (rMajor == null) rMajor = sequence(1, 16);
    if (rMinor == null) rMinor = sequence(1, 15.5, by: 0.5);
    List<String> 
      rMajorcolorList = _extractStrList(rMajorcolor),
      rMinorcolorList = _extractStrList(rMinorcolor)
    ;
    List<num>
      rMajorList = _extractNumList(rMajor),
      rMinorList = _extractNumList(rMinor),
      rMajorWidthList = _extractNumList(rMajorWidth),
      rMinorWidthList = _extractNumList(rMinorWidth)
    ;
    
    for (int i = 0; i < rMinorList.length; i ++) curveParametric(
      x: (t) => rMinorList[i] * Math.cos(t),
      y: (t) => rMinorList[i] * Math.sin(t),
      color: rMinorcolorList[i % rMinorcolorList.length],
      width: rMinorWidthList[i % rMinorWidthList.length]
    );
    
    for (int i = 0; i < rMajorList.length; i ++) curveParametric(
      x: (t) => rMajorList[i] * Math.cos(t),
      y: (t) => rMajorList[i] * Math.sin(t),
      color: rMajorcolorList[i % rMajorcolorList.length],
      width: rMajorWidthList[i % rMajorWidthList.length]
    );
    
    (new List<num>.generate(
      5, 
      (int i) => Math.tan(Math.PI / 12 * (i + 1))
    )).forEach((num m) {
      curve(
        model: (x) => m * x,
        color: "rgba(25,200,75,0.1)",
        width: 1
      );
      curve(
        model: (x) => -m * x,
        color: "rgba(25,200,75,0.1)",
        width: 1
      );
    });
    
    linesVertical(x: 0, color: "rgba(0,0,0,0.4)", width: 1);
    linesHorizontal(y: 0, color: "rgba(0,0,0,0.4)", width: 1);
  }
  
/// Adds a curve to the plot.
/// 
/// The parameter [model] is either a [Function] or a [Model]; [dx] 
/// determines the "step size" and smaller values may result in smoother 
/// curves. Asymptotes can be added via the parameters [asymptotesVertical] 
/// and [asymptotesHorizontal] to produce better looking plots if applicable.
/// The [bool] [inDrawRegion] determines whether the curve will only be 
/// visible in the draw region determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.curve(
///       model: (x) => 3 / (x - 3) + 2,
///       asymptotesVertical: 3,
///       asymptotesHorizontal: 2,
///       dx: 0.01
///     );
///     
/// ...
  curve({
    var model, 
    num dx: 0.1,
    var asymptotesVertical: const [],
    String asymptoteVerticalcolor: "rgba(200, 50, 75, 0.3)",
    num asymptoteVerticalWidth: 1,
    var asymptotesHorizontal: const [],
    String asymptoteHorizontalcolor: "rgba(200, 50, 75, 0.3)",
    num asymptoteHorizontalWidth: 1,
    String color: "rgba(50, 75, 200, 0.8)",
    num width: 2,
    bool inDrawRegion: true,
    List<num> lineStyle
  }) {
    Function f = model is Function ? model : model.f;
    List<num> 
      asymptotesVerticalData = _extractNumList(asymptotesVertical),
      asymptotesHorizontalData = _extractNumList(asymptotesHorizontal)
    ;
    var 
      x = new List<num>.generate(
        ((xRange[1] - xRange[0] + 2) / dx).round(), 
        (int i) => xRange[0] - 1 + i * dx
      )
    ;
    
    if (asymptotesVerticalData.isNotEmpty) {
      x = x.where((num a) => !asymptotesVerticalData.contains(a)).toList();
      for (int i = 0; i < asymptotesVerticalData.length; i++) {
        List<num> xTake = x.takeWhile(
          (num x0) => x0 < asymptotesVerticalData[i]).toList();
        x = x.sublist(xTake.length);
        List<num> y = xTake.map(f).toList();
        lines(
          x: xTake, 
          y: y,
          width: width,
          color: color,
          inDrawRegion: inDrawRegion
        );
        
      }
      linesVertical(
        x: asymptotesVerticalData, 
        color: asymptoteVerticalcolor,
        width: asymptoteVerticalWidth,
        inDrawRegion: inDrawRegion
      );
      
    } 
    
    List<num> y = x.map(f).toList();
    lines(
      x: x, 
      y: y,
      width: width,
      color: color,
      inDrawRegion: inDrawRegion
    );
    
    
    if (asymptotesHorizontalData.isNotEmpty) {
      linesHorizontal(
        y: asymptotesHorizontalData, 
        color: asymptoteHorizontalcolor,
        width: asymptoteHorizontalWidth,
        inDrawRegion: inDrawRegion
      ); 
    }
  }
  
/// Adds a parametric curve to the plot.
/// 
/// The arguments [x] and [y] are each either a [Function] or a [Model] 
/// expressed in terms of some [num] parameter; [tInit], [tFinal] and [dt]
/// determine the initial, final and "step size" of the parameter; 
/// [inDrawRegion] determines whether the curve will only be visible in 
/// the draw region determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
///
///     var plot = new Plot();
///     plot.curveParametric(
///       x: (t) => 8 * cos(3 * t) * cos(t),
///       y: (t) => 8 * cos(3 * t) * sin(t)
///     );
///     
/// ...
  curveParametric({
    var x,
    var y,
    num tInit: 0,
    num tFinal: Math.PI * 2,
    num dt: Math.PI / 128,
    String color: "rgba(50, 75, 200, 0.8)",
    num width: 2,
    bool inDrawRegion: true
  }) {
    Function 
      fx = x is Function ? x : x.f,
      fy = y is Function ? y : y.f
    ;
    
    List<num>
      t = new List<num>.generate(
        ((tFinal - tInit) / dt).round() + 1, 
        (int i) => tInit + i * dt
      ),
      xData = t.map(fx).toList(),
      yData = t.map(fy).toList()
    ;
    
    lines(
      x: xData,
      y: yData,
      color: color,
      width: width,
      inDrawRegion: inDrawRegion
    );
  }
  
/// Adds a polar curve to the plot.
/// 
/// The argument [r] is each either a [Function] or a [Model] expressed 
/// in terms of some [num] parameter; [tInit], [tFinal] and [dt] determine 
/// the initial, final and "step size" of the parameter; [inDrawRegion] 
/// determines whether the curve will only be visible in the draw region 
/// determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.curvePolar(
///       r: (t) => 9 * cos(6 / 5 * t),
///       color: "rgba(255, 50, 125, 0.6)",
///       width: 5,
///       tFinal: 10 * PI
///     );
///     
/// ...
  curvePolar({
    var r,
    num tInit: 0,
    num tFinal: Math.PI * 2,
    num dt: Math.PI / 128,
    String color: "rgba(50, 75, 200, 0.8)",
    num width: 2,
    bool inDrawRegion: true
  }) {
    Function fr = r is Function ? r: r.f;
    List<num> 
      theta = sequence(tInit, tFinal, by: dt, inclusive: true),
      rList = theta.map(fr).toList(),
      x = new List<num>.generate(
        theta.length, 
        (int i) => rList[i] * Math.cos(theta[i])
      ),
      y = new List<num>.generate(
        theta.length, 
        (int i) => rList[i] * Math.sin(theta[i])
      )
    ;
    lines(
      x: x,
      y: y,
      color: color,
      width: width,
      inDrawRegion: inDrawRegion
    );
      
  }

  
// TODO: fix!
/// Adds a curve to the plot from a gradient function and a point.
/// 
/// The argument [m] is each either a [Function] or a [Model] expressed 
/// in terms of two parameters representing the x and y coordinates; 
/// [x] and [y] are to hold the point that the curve is constrained to
/// pass through; [inDrawRegion] determines whether the curve will only 
/// be visible in the draw region determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.curveDifferential(
///       m: (x, y) => x + y , 
///       x: 0, 
///       y: [-9, -7, -5, -3, 0, 3, 5]
///     );
///     
/// ...
  
  curveDifferential({
    var m,
    var x,
    var y,
    num r: 0.01,
    var color: "rgba(0,0,255,0.5)",
    var width: 2,
    bool inDrawRegion: true
  }) {
    Function gradient = m is Function ? m : m.f;
    List<num> 
      xList = _extractNumList(x), 
      yList = _extractNumList(y),
      widthList = _extractNumList(width)
    ;
      
    int length = Math.max(xList.length, yList.length);
    List<String> colorList = _extractStrList(color);
    
    for (int i = 0; i < length; i++) {
      num x0 = xList[i % xList.length], y0 = yList[i % yList.length];
      var f = (num d) {
        num xPlot1 = x0, yPlot1 = y0, xPlot2, yPlot2;
        List<num> Xs = [xPlot1], Ys = [yPlot1];
        num yAccept = 10 * (yRange[1] - yRange[0]);
        //num thetaPrev;
        while (
          (xRange[0] < xPlot1 && xPlot1 < xRange[1])
          && (yRange[0] - yAccept < yPlot1 && yPlot1 < yRange[0] + yAccept)
        ) {
          num gr = (/*xPlot1 == 0 && */yPlot1 == 0) ? Math.PI / 2 : gradient(xPlot1,  yPlot1);

          num theta = Math.atan2(gr, 1);
          xPlot2 = xPlot1 + r * Math.cos(theta + d * Math.PI);
          yPlot2 = yPlot1 + r * Math.sin(theta + d * Math.PI);
          xPlot1 = xPlot2;
          yPlot1 = yPlot2;
          Xs.add(xPlot1); Ys.add(yPlot1);
        }
        lines(
          x: Xs, 
          y: Ys, 
          color: colorList[i % colorList.length],
          width: widthList[i % widthList.length]
        );
        points(x: x0, y: y0);
      };
      f(0);
      f(1);
      
    }
  }

/// Draws a gradient field defined by [m] at points with 
/// coordinates [x] and [y].
  gradientField({
    var m,
    var x, 
    var y,
    String color: "rgba(0,0,0,0.3)",
    num width: 1,
    bool inDrawRegion: true
  }) {
    Function gradient = m is Function ? m : m.f;
    num len = Math.min(
      (xRange[1] - xRange[0]) / 40, 
      (yRange[1] - yRange[0]) / 40
    );
    if (x == null) x = new List.generate(
      25, 
      (int i) => xRange[0] + (xRange[1] - xRange[0]) / 20 * i
    );
    if (y == null) y = new List.generate(
      25, 
      (int i) => yRange[0] + (yRange[1] - yRange[0]) / 20 * i
    );
    List<num> xList = _extractNumList(x), yList = _extractNumList(y);
    xList.forEach((num x0) {
      yList.forEach((num y0) {
        num 
          theta = Math.atan2(gradient(x0, y0), 1),
          dx = len * Math.cos(theta),
          dy = len * Math.sin(theta),
          x1 = x0 - dx, x2 = x0 + dx,
          y1 = y0 - dy, y2 = y0 + dy
        ;
        lines(x: [x1, x2], y: [y1, y2], color: color, width: width);
      });
    });
  }
  
/// Adds points to the plot. 
/// 
/// The arguments [x], [y], [lineWidth] and [size] can be [num]s, 
/// [NumericVector]s or [List]s, in which latter two cases they will be 
/// iterated along with the coordinates; similarly [type], [fillColor] and 
/// [strokeColor] can be [String]s, [CategoricVector]s or [List]s.
/// 
/// The argument [type] is a character to be plotted or a value stored 
/// in the class [PointType].
/// 
/// The argument [inDrawRegion] determines whether the curve will only be 
/// visible in the draw region determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.points(
///       x: [0, 3, 6],
///       y: [5, 4, 3],
///       type: PointType.DIAMOND
///     ); 
///     
/// ...
  points({
    var x, 
    var y,
    var type: PointType.CIRCLE,
    var fillColor: "rgba(0,0,0,0.2)", //"rgba(100,125,255,0.3)",
    var strokeColor: "rgba(0,0,0,0.4)",
    var lineWidth: 1,
    var size,
    bool inDrawRegion: true
  }) {
    if (size == null) size = canvas.width / 180;
    List<num> 
      xData = _extractNumList(x),
      yData = _extractNumList(y),
      pointLineWidthData = _extractNumList(lineWidth),
      pointSizeData = _extractNumList(size)
    ; 
      
    List<String> 
      pointTypeData = _extractStrList(type),
      pointcolorData = _extractStrList(fillColor),
      strokecolorData = _extractStrList(strokeColor)
    ;
    
    _drawRegion(inDrawRegion);
    
    int length = Math.max(xData.length, yData.length);
    for (int i = 0; i < length; i++) {
      num 
        x0 = _transformX(xData[i % xData.length]),
        y0 = _transformY(yData[i % yData.length]),
        p0 = pointSizeData[i % pointSizeData.length]
      ;
      layerContext
        ..fillStyle = pointcolorData[i % pointcolorData.length]
        ..strokeStyle = strokecolorData[i % strokecolorData.length]
        ..lineWidth = pointLineWidthData[i % pointLineWidthData.length]
      ;
      String char = pointTypeData[i % pointTypeData.length];
      switch (char) {
        case "_o":
          layerContext
            ..beginPath()
            ..arc(
                x0, y0, 
                p0, 
                0, Math.PI * 2
              )
            ..fill()
            ..stroke()
            ..closePath()
          ;
          break;
        case "_0":
          num r = 0.9;
          layerContext
            ..beginPath()
            ..moveTo(x0 - p0 * r, y0 - p0 * r)
            ..lineTo(x0 + p0 * r, y0 - p0 * r)
            ..lineTo(x0 + p0 * r, y0 + p0 * r)
            ..lineTo(x0 - p0 * r, y0 + p0 * r)
            ..lineTo(x0 - p0 * r, y0 - p0 * r)
            ..fill()
            ..stroke()
            ..closePath()
          ;
          break;
        case "_D":
          num r = 1.2;
          layerContext
            ..beginPath()
            ..moveTo(x0, y0 - p0 * r)
            ..lineTo(x0 + p0 * r, y0)
            ..lineTo(x0, y0 + p0 * r)
            ..lineTo(x0 - p0 * r, y0)
            ..lineTo(x0, y0 - p0 * r)
            ..fill()
            ..stroke()
            ..closePath()
          ;
          break;
        case "_^":
          layerContext
            ..beginPath()
            ..moveTo(x0, y0 - p0)
            ..lineTo(x0 + p0, y0 + p0)
            ..lineTo(x0 - p0, y0 + p0)
            ..lineTo(x0, y0 - p0)
            ..fill()
            ..stroke()
            ..closePath()
          ;
          break;
        case "_<":
          layerContext
            ..beginPath()
            ..moveTo(x0 - p0, y0)
            ..lineTo(x0 + p0, y0 - p0)
            ..lineTo(x0 + p0, y0 + p0)
            ..lineTo(x0 - p0, y0)
            ..fill()
            ..stroke()
            ..closePath()
          ;
          break;
        case "_>":
          layerContext
            ..beginPath()
            ..moveTo(x0 - p0, y0 - p0)
            ..lineTo(x0 + p0, y0)
            ..lineTo(x0 - p0, y0 + p0)
            ..lineTo(x0 - p0, y0 - p0)
            ..fill()
            ..stroke()
            ..closePath()
          ;
          break;
        case "_v":
          layerContext
            ..beginPath()
            ..moveTo(x0 - p0, y0 - p0)
            ..lineTo(x0 + p0, y0 - p0)
            ..lineTo(x0, y0 + p0)
            ..lineTo(x0 - p0, y0 - p0)
            ..fill()
            ..stroke()
            ..closePath()
          ;
          break;
        case "_+":
          layerContext
            ..beginPath()
            ..moveTo(x0, y0 - p0)
            ..lineTo(x0, y0 + p0)
            ..moveTo(x0 - p0, y0)
            ..lineTo(x0 + p0, y0)
            ..stroke()
            ..closePath()
          ;
          break;
        case "_x":
          num r = 0.8;
          layerContext
            ..beginPath()
            ..moveTo(x0 - p0 * r, y0 - p0 * r)
            ..lineTo(x0 + p0 * r, y0 + p0 * r)
            ..moveTo(x0 - p0 * r, y0 + p0 * r)
            ..lineTo(x0 + p0 * r, y0 - p0 * r)
            ..stroke()
            ..closePath()
          ;
          break;
        default:
          char = char[0];
          layerContext
            .font = "${(p0 * 3).round()}px Courier";
          num d = layerContext
            .measureText(char).width / 2;
          layerContext
            ..fillText(
                char, 
                x0 - d, 
                y0 + 0.75 * d
              )
            ..strokeText(
                char, 
                x0 - d, 
                y0 + 0.75 * d
              )
          ;
          break;
      }
    }
    
    bufferContext.drawImage(layer, 0, 0);
    
    context.drawImage(buffer, 0, 0);
    
  }
  
/// Adds a polygon to the plot. 
/// 
/// The arguments [x], [y], [lineWidth] should be [num]s, [NumericVector]s 
/// or [List]s. Similarly, [lineColor] and [fillColor] should be [String]s, 
/// [CategoricVector]s or [List]s. 
/// 
/// The argument [inDrawRegion] determines whether the curve will only be
/// visible in the draw region determined by [drawXRange] and [drawYRange].
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.polygon(
///       x: [0, 0, 5, 5],
///       y: [0, 8, 5, 0]
///     );
///     
/// ...
  polygon({
    var x,
    var y,
    var lineColor: "blue",
    var lineWidth: 1,
    String fillColor: "rgba(0,0,255,0.2)",
    //String shadow: "none",
    bool inDrawRegion: true
  }) {
    List<num>
      xData = _extractNumList(x),
      yData = _extractNumList(y),
      lineWidthData = _extractNumList(lineWidth)
    ;
    
    _drawRegion(inDrawRegion);
      
    layerContext
      ..fillStyle = fillColor
      ..beginPath()
      ..moveTo(_transformX(xData[0]), _transformY(yData[0]))
    ;
    
    // Fill
    int length = Math.max(xData.length, yData.length);
    for (int i = 1; i < length; i++) {
      layerContext
        ..lineTo(
          _transformX(xData[i % xData.length]), 
          _transformY(yData[i % yData.length])
        )
      ;
    }
    layerContext
      ..closePath()
      ..fill()
    ;
    
    bufferContext.drawImage(layer, 0, 0);
    
    context.drawImage(buffer, 0, 0);
    
    lines(
      x: xData..add(xData[0]), 
      y: yData..add(yData[0]), 
      color: lineColor, 
      width: lineWidth
    );
  }
  
/// Adds a generic scatter-plot to the plot. 
/// 
/// The arguments [x], [y], should contain the coordinates and be of type 
/// [num], [NumericVector] or [List]. The title, x-axis and y-axis labels 
/// are specified via [title], [xLabel] and [yLabel] respectively. If 
/// [plotPoints] is false, the points will not be plotted (for example, if 
/// the user only wants to set up the graph scale and labels and will plot 
/// points using method [points] later).
/// 
/// ...
/// Example:
/// 
///     var plot = new Plot();
///     plot.scatter(
///       x: [0, 0, 5, 5],
///       y: [0, 8, 5, 0],
///       title: "Title",
///       xLabel: "X",
///       yLabel: "Y"
///     );
///     
/// ...
  scatter({
    var x,
    var y,
    bool plotPoints: true,
    String xLabel,
    String yLabel,
    String title,
    bool summary: false,
    bool border: true
  }) {
    NumericVector 
      xVec = x is NumericVector ? x : 
        new NumericVector(data: _extractNumList(x)),
      yVec = y is NumericVector ? y : 
        new NumericVector(data: _extractNumList(y))
    ;
    
    _setUpPlotRegion(
      minX: xVec.min, 
      maxX: xVec.max, 
      minY: yVec.min, 
      maxY: yVec.max,
      title: title,
      xLabel: xLabel,
      yLabel: yLabel,
      markX: true,
      markY: true,
      border: border,
      summary: summary
    );
    
    // points
    if (plotPoints) points(
      x: xVec,
      y: yVec
    );
    
    // Means
    linesVertical(x: xVec.mean, color: "rgba(0,0,0,0.1)");
    linesHorizontal(y: yVec.mean, color: "rgba(0,0,0,0.1)");
  }
  
/// Adds a generic box-plot to the plot. 
/// 
/// The arguments [data] specifies the data to be summarised. If [categories]
/// is specified, box-plots will be produced by category.
  boxPlot({
    var data,
    var categories,
    bool horizontal: false,
    String xLabel,
    String yLabel,
    String title,
    bool summary: false
  }) {
    var dataVec = new NumericVector(data: _extractNumList(data));
    var catVec = categories == null ? 
      new CategoricVector(data: new List<bool>.filled(dataVec.count, true)) :
        new CategoricVector(data: _extractStrList(categories));
      
      
    _setUpPlotRegion(
      minX: horizontal ? dataVec.min : 0, 
      maxX: horizontal ? dataVec.max : catVec.categories.length, 
      minY: horizontal ? 0 : dataVec.min, 
      maxY: horizontal ? catVec.categories.length : dataVec.max,
      title: title,
      xLabel: xLabel,
      yLabel: yLabel,
      markX: horizontal,
      markY: !horizontal,
      summary: summary
    );
      
    num i = 0.5;
      for (String category in catVec.categories) {
        NumericVector v = dataVec[catVec == category];
          
        // background
        polygon(
          x: horizontal ? 
            [drawXRange[0], drawXRange[0], drawXRange[1], drawXRange[1]] : 
              [i - 0.3, i + 0.3, i + 0.3, i - 0.3],
          y: horizontal ? 
            [i - 0.3, i + 0.3, i + 0.3, i - 0.3] : 
              [drawYRange[0], drawYRange[0], drawYRange[1], drawYRange[1]],
          fillColor: "rgba(0, 0, 0, 0.05)",
          lineColor: "rgba(0,0,0,0)"
        );
          
        // box
        polygon(
          x: horizontal ? 
            [v.q1, v.q1, v.q3, v.q3] : 
              [i - 0.45, i + 0.45, i + 0.45, i - 0.45],
          y: horizontal ? 
            [i - 0.45, i + 0.45, i + 0.45, i - 0.45] : 
              [v.q1, v.q1, v.q3, v.q3],
          fillColor: "rgb(245,245,245)",
            lineColor: "black",
            lineWidth: canvas.width / 300
        );
          
        // median
        lines(
          x: horizontal ? [v.median, v.median] : [i - 0.4, i + 0.4],
          y: horizontal ? [i - 0.4, i + 0.4] : [v.median, v.median],
          color: "rgba(0,0,0,0.8)"
        );
          
        num 
          iqr = v.q3 - v.q1,
          cutOff = iqr * 1.5
        ;
          
        NumericVector 
          vLow = v[v <= v.q3 + cutOff], 
          vHigh = v[v >= v.q1 - cutOff]
        ;
          
          
        // upper quartile
        lines(
          x: horizontal ? [v.q3, vLow.max] : [i, i],
          y: horizontal ? [i, i] : [v.q3, vLow.max],
          color: "rgba(0,0,0,0.8)",
          width: canvas.width / 300
        );
          
        // lower quartile
        lines(
          x: horizontal ? [v.q1, vHigh.min] : [i, i],
          y: horizontal ? [i, i] : [v.q1, vHigh.min],
          color: "rgba(0,0,0,0.8)",
          width: canvas.width / 300
        );
          
        // upper outliers
        if (vHigh.any((x) => x > v.q3 + cutOff)) points(
          x: horizontal ? vHigh[vHigh > v.q3 + cutOff] : i,
          y: horizontal ? i : vHigh[vHigh > v.q3 + cutOff],
          fillColor: "rgba(0,0,0,0)",
          strokeColor: "rgba(0,0,0,0.8)"
        );
          
        // lower outliers
        if (vLow.any((x) => x < v.q1 - cutOff)) points(
          x: horizontal ? vLow[vLow < v.q1 - cutOff] : i,
          y: horizontal ? i : vLow[vLow < v.q1 - cutOff],
          fillColor: "rgba(0,0,0,0)",
          strokeColor: "rgba(0,0,0,0.8)"
        );
          
        // category labels
        labels(
          x: horizontal ? 
              _transformX(drawXRange[0]) - canvas.width / 50 : 
              _transformX(i),
          y: horizontal ? 
              _transformY(i) : 
              _transformY(drawYRange[0]) + canvas.width / 60,
          labels: category,
          theta: horizontal ? Math.PI / 2 : 0,
          color: "gray",
          scaled: false  
        );
          
        i++;
        
    }   
  }

/// Adds a generic mosaic-plot to the plot. 
/// 
/// The arguments [categoriesX] and [categoriesY] specify the data to be 
/// summarised, expected to be of type [CategoricVector], [String] or [List]. 
/// If [freq] is specified it is taken to represent the frequency in each
/// row of the data.
  mosaic({
    var categoriesX,
    var categoriesY,
    var freq,
    String xLabel,
    String yLabel,
    String title,
    bool summary: false
  }) {
    CategoricVector 
      catXVec = new CategoricVector(data: _extractStrList(categoriesX)),
      catYVec = new CategoricVector(data: _extractStrList(categoriesY))
    ;
    int length = Math.max(catXVec.count, catYVec.count);
    NumericVector freqVec = freq == null ? 
      new NumericVector(data: new List<num>.filled(length, 1)) :
        new NumericVector(data: _extractNumList(freq));
    
    
    _setUpPlotRegion(
      minX: 0, 
      maxX: catXVec.categories.length, 
      minY: 0, 
      maxY: catYVec.categories.length,
      title: title,
      xLabel: xLabel,
      yLabel: yLabel,
      markX: false,
      markY: false,
      summary: summary
    );
    
    num xLeft = 0.5;
    //int top = 0;
    for (String catX in catXVec.categories) {
      NumericVector xVec = freqVec[catXVec == catX];
      num xRight = xLeft 
        + xVec.sum / freqVec.sum * catXVec.categories.length;
        
      num yBottom = 0.5;
      int darken = 0;
      for (String catY in catYVec.categories) {
        NumericVector conjunction = freqVec[
          ((catXVec == catX) as BooleanVector) 
          & ((catYVec == catY) as BooleanVector)
        ];
          
        num yTop = yBottom 
          + conjunction.sum / xVec.sum * catYVec.categories.length;
          
        polygon(
          x: [xLeft - 0.5, xLeft - 0.5, xRight - 0.5, xRight - 0.5],
          y: [yBottom - 0.5, yTop - 0.5, yTop - 0.5, yBottom - 0.5],
          fillColor: darken % 2 == 0 ? 
            "rgba(0,0,0,0.05)" : "rgba(0,0,0,0.15)",
          lineColor: "black",
          lineWidth: canvas.width / 200
        );
        labels(
          x: (xLeft + xRight - 1) / 2,
          y: (yBottom + yTop - 1) / 2,
          color: "black",
          labels: catY
        );
          
          
        yBottom = yTop;
        darken ++;
              
      }
      num tx = _transformX((xLeft + xRight - 1) / 2);
          //ty = _transformY(drawYRange[1]) + canvas.height / 40;
      labels(
        x: [tx, tx],
        y: [_transformY(drawYRange[1]) + canvas.height / 35,
          _transformY(drawYRange[0]) - canvas.height / 30],
        labels: catX,
        color: "gray",
        scaled: false
      );
      xLeft = xRight;
     // top++;
    }
    
         
  }
  
/// Adds a generic histogram to the plot.
/// 
/// Expects [data] to be a [NumericVector] or a [List].
  histogram({
    var data,
    String xLabel,
    String yLabel,
    String title,
    int bins,
    bool summary: false
  }) {
    NumericVector dataVec = data is NumericVector ? data : 
      new NumericVector(data: _extractNumList(data));
    
    if (bins == null) bins = Math.sqrt(dataVec.count).round();
    num binSize = (dataVec.max - dataVec.min) / bins;
    List<num> binBounds = new List<num>.generate(
      bins, 
      (int i) => dataVec.min + i * binSize 
    )..add(dataVec.min + bins * binSize);
    
    List<num> counts = new List<num>.generate(
      bins,
      (int i) => ((dataVec >= binBounds[i]) & (dataVec < binBounds[i + 1])).sum
    );
    
    NumericVector
      xVec = new NumericVector(data: binBounds),
      yVec = new NumericVector(data: counts)
    ;
    
    _setUpPlotRegion(
      minX: xVec.min,
      maxX: xVec.max,
      minY: 0,
      maxY: yVec.max,
      xLabel: xLabel,
      yLabel: yLabel,
      title: title,
      markX: true,
      markY: true,
      border: true,
      summary: summary
    );
        
    
    for (int i = 0; i < yVec.count; i++) {
      polygon(
        x: [xVec[i], xVec[i], xVec[i] + binSize, xVec[i] + binSize],
        y: [0, yVec[i], yVec[i], 0],
        fillColor: "rgb(245,245,245)",
        lineColor: "rgba(0,0,0,0)"
      );
      lines(
        x: [xVec[i], xVec[i] + binSize],
        y: yVec[i],
        color: "black",
        width: canvas.width / 500
      );
      num height = i == 0 ? yVec[i] : Math.max(yVec[i], yVec[i - 1]);
      lines(
        x: [xVec[i], xVec[i]], 
        y: [0, height], 
        color: "black",
        width: canvas.width / 500
      );
      
    }
    lines(
      x: [xVec.last, xVec.last], 
      y: [0, yVec.last],
      color: "black",
      width: canvas.width / 500
    );
    linesHorizontal(
      y: 0,
      color: "black",
      width: canvas.width / 300
    );

     
  }

/// Adds a generic bar-chart to the plot.
/// 
/// Expects [categories] to be a [CategoricVector] of a [List]. If [frequency] is specified it is considered as the frequency associated
/// with each row.
  barChart({
    var categories,
    var frequency,
    String xLabel,
    String yLabel,
    String title,
    bool summary: false
  }) {
    CategoricVector catVec = categories is CategoricVector ? categories : 
      new CategoricVector(data: _extractStrList(categories));
    
    NumericVector freqVec = frequency == null ? 
      new NumericVector(data: new List<num>.filled(catVec.count, 1)) :
        new NumericVector(data: _extractNumList(frequency));
    
    NumericVector countVec = new NumericVector(
      data: catVec.categories.map(
        (String cat) => freqVec[catVec == cat].sum / freqVec.sum
      ).toList()
    );
      
    _setUpPlotRegion(
      minX: 0,
      maxX: catVec.categories.length,
      minY: 0,
      maxY: countVec.max,
      xLabel: xLabel,
      yLabel: yLabel,
      title: title,
      markX: false,
      markY: true,
      border: true,
      summary: summary
    );   
      
    for (num i = 0.5; i < countVec.count; i++) {
      int index = i.floor();
      polygon(
        x: [i - 0.4, i - 0.4, i + 0.4, i + 0.4],
        y: [0, countVec[index], countVec[index], 0],
        fillColor: "rgb(245,245,245)",
        lineColor: "rgba(0,0,0,0)"
      );
      lines(
        x: [i - 0.4, i + 0.4],
        y: countVec[index],
        color: "black",
        width: canvas.width / 500
      );
      num height = i == 0 ? 
        countVec[index] : 
          Math.max(countVec[index], countVec[index - 1]);
      
      lines(
        x: [i - 0.4, i - 0.4], 
        y: [0, height], 
        color: "black",
        width: canvas.width / 300
      );
      lines(
        x: [i + 0.4, i + 0.4], 
        y: [0, height], 
        color: "black",
        width: canvas.width / 300
      );
      
      // category labels
      labels(
        x: _transformX(i),
        y: _transformY(drawYRange[0]) + canvas.width / 60,
        labels: catVec.categories[index],
        color: "gray",
        scaled: false  
      );
    }
    
    linesHorizontal(
      y: 0,
      color: "black",
      width: canvas.width / 300
    );
       
  }
  
}